import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.io.*;
import java.util.StringTokenizer;
import java.text.DecimalFormat;

public class VDatum extends JFrame {

//*** program to convert vertical datums

//*** internal units of meters and decimal degrees, long positive EAST
//*** input lat/lon's converted immediately to the above

//*** batch mode delimiters are " \t\n\r\f,;:" 
//***  (space,tab,newline,carriage-return,form-feed,comma,semicolon,colon)

//*** batch mode allows a "lead minus" longitude form  (e.g. -82.484) 

//***note: currently two button handler instances (same object only need 1)
//***note: more sophisticated processing of last field (batch mode)
//*** 23-aug-00  fixed sign error on application of tide datums
//*** 30-sep-00  added create file capability to output file chooser code
//*** 26-jul-01  use blterp for sst (to allow for null values)
//*** 23-dec-01  added itrf00 frame
//*** 25-jan-02  fixed sign error for sst when going from 3-d to tidal
//*** 16-sep-02  added wgs84(G1150) frame

   public static final String version = "V-Datum 1.06";

   public static final 
           String hdatS[] = {"NAD 83, WGS, ITRF",      //***  0
                             "NAD 27"};                //***  1

   public static final 
           String vdatS[] = {"NAVD 88",                //***  0  <--- ORTIDX
                             "NGVD 29",                //***  1
                             "MLLW",                   //***  2  <--- TIDIDX
                             "MLW",                    //***  3
                             "LMSL",                   //***  4  <--- MSLIDX
                             "MTL",                    //***  5
                             "DTL",                    //***  6
                             "MHW",                    //***  7
                             "MHHW",                   //***  8
                             "NAD 83 (86)",            //***  9  <--- NADIDX
                             "WGS 84(G1150)",          //*** 10
                             "WGS 84(G873)",           //*** 11
                             "WGS 84(G730)",           //*** 12
                             "WGS 84(orig.)",          //*** 13
                             "WGS 72",                 //*** 14
                             "ITRF2000(1997.0) ",      //*** 15
                             "ITRF97(1997.0)   ",      //*** 16
                             "ITRF96(1997.0)   ",      //*** 17
                             "ITRF94",                 //*** 18
                             "ITRF93",                 //*** 19
                             "ITRF92",                 //*** 20
                             "ITRF91",                 //*** 21
                             "ITRF90",                 //*** 22
                             "ITRF89",                 //*** 23
                             "ITRF88",                 //*** 24
                             "SIO/MIT 92",             //*** 25
                             "NEOS 90",                //*** 26
                             "PNEOS 90"};              //*** 27

   public static final int ORTIDX = 0, TIDIDX = 2, MSLIDX = 4, NADIDX = 9;
   public TSet frameTS[];              //*** frame transformations[9-->27]
   public Grid nclaG,ncloG,g99G,vcnG,sstG;
   public Grid mllwG,mlwG,mhwG,mhhwG,mtlG,dtlG;
   public static final DecimalFormat dec4DF=new DecimalFormat("#0.0000");

   private BufferedReader nclaBR,ncloBR,g99BR,vcnBR,sstBR;   //*** grid files
   private BufferedReader mllwBR,mlwBR,mhwBR,mhhwBR;         //*** grid files
   private BufferedReader mtlBR,dtlBR;                       //*** grid files
   private File   inFile,outFile;    //******debug  (global?)

   private JLabel lab22;                      //*** used for error messages
   private JTextField latTF, lonTF, hinTF, houtTF;
   private JComboBox  hdatCB,   vindatCB, voutdatCB;
   private int        horIdx=0, vinIdx=0, voutIdx=0;

   private JButton               intrButtn, batButtn, finButtn, foutButtn;
   private JRadioButtonMenuItem  modeItems[];
   private ButtonGroup           modeGroup;
   private int                   iMode=0;     //*** default mode = Interactive

   private JPanel     lhPanel,lh0Panel,lh1Panel,rhPanel;
   private CardLayout myCL;
   private Color bgColor = new Color(220,220,255);            //*** light blue
   private Font labFont = new Font("Monospaced",Font.PLAIN,14);    //*** label
   private JTextField finTF, foutTF;                  //*** batch file names
   private BufferedReader finBR;                             //*** batch input
   private BufferedWriter foutBR;                            //*** batch output

   public static final double TOFEET = 3937.0/1200.0;      //*** meters to feet
   private JRadioButton       feetRButtn, meterRButtn;
   private ButtonGroup        fmRBGroup;
   private int                iFeet=0;             //*** default units = Meters

   private JRadioButton       soundRButtn, htRButtn;
   private ButtonGroup        hsRBGroup;
   private int                iSound=0;            //*** default data = Heights

   private JRadioButton       lonLatRButtn, latLonRButtn;
   private ButtonGroup        yxRBGroup;
   private int                iLonLat=0;           //*** default data = lat,lon

   public static final double DNULL = -999999.0;  //*** null value

//***
//*** end globals
//***

   public static void main(String args[]) { 
      VDatum app = new VDatum();

      app.addWindowListener(
         new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
               System.exit(0);
            }
         }
      );
   }

   public VDatum() {
      super("Vertical Datum Transformation");

//*** windows look and feel

      try {
         UIManager.setLookAndFeel(
                  "com.sun.java.swing.plaf.windows.WindowsLookAndFeel");
      } catch (InstantiationException e) {
                             Dump.Prog("LookAndFeel error 1");
      } catch (ClassNotFoundException e) {
                             Dump.Prog("LookAndFeel error 2");
      } catch (UnsupportedLookAndFeelException e) {
                             Dump.Prog("LookAndFeel error 3");
      } catch (IllegalAccessException e) {
                             Dump.Prog("LookAndFeel error 4");
      }

//*** menu section

      JMenuBar myMB = new JMenuBar();   //*** create menubar
      setJMenuBar(myMB);                //*** set the menubar for the JFrame

//*** create the File menu

      JMenu fileMenu = new JMenu("File");
      fileMenu.setMnemonic('F');

      JMenuItem aboutItem = new JMenuItem("About...");
      aboutItem.setMnemonic('A');
      aboutItem.addActionListener(
         new ActionListener() {
            public void actionPerformed(ActionEvent e) {
               JOptionPane.showMessageDialog(VDatum.this,
                  "Vertical Datum Transformation\n"+
                  "        " + version,
                  "About", JOptionPane.PLAIN_MESSAGE);
            }
         }
      );
      fileMenu.add(aboutItem);

      JMenuItem exitItem = new JMenuItem("Exit");
      exitItem.setMnemonic('x');
      exitItem.addActionListener(
         new ActionListener() {
            public void actionPerformed(ActionEvent e) {System.exit(0);}
         }
      );
      fileMenu.add(exitItem);

      myMB.add(fileMenu);                    //*** add File menu to menu bar

//*** create the Mode menu   (menu radio button selections)

      JMenu modeMenu = new JMenu("Mode");  
      modeMenu.setMnemonic('M');

      String    modes[] = {"Interactive", "Batch"};
      char   memModes[] = {'I',           'B'};
      modeItems = new JRadioButtonMenuItem[modes.length];
      modeGroup = new ButtonGroup();
      ItemHandler modeHandler = new ItemHandler();

      for (int i=0; i<modes.length; i++) {
         modeItems[i] = new JRadioButtonMenuItem(modes[i]);
         modeMenu.add (modeItems[i]);
         modeGroup.add(modeItems[i]);
         modeItems[i].setMnemonic(memModes[i]);
         modeItems[i].addActionListener(modeHandler);
      }
      modeItems[iMode].setSelected(true);    //*** default mode = Interactive

      myMB.add(modeMenu);                    //*** add Mode menu to menu bar

//*** end menu -- now for GUI layout

      Container c = getContentPane();
      c.setLayout(new FlowLayout());          //*** flow layout (outermost)
      c.setBackground(bgColor);

//*** basically, left and right halves  (6 rows, 1 col each)
//*** two different version of left half, held in a card layout

      lhPanel = new JPanel(new CardLayout());         //*** holder

      lh1Panel = new JPanel();         //*** batch
      lh0Panel = new JPanel();         //*** interactive

      leftPanel1(lh1Panel);            //*** build batch panel
      leftPanel0(lh0Panel);            //*** build interactive panel

//*** build right half

      Font mesFont = new Font ("SansSerif", Font.BOLD, 14);  //*** message font

      rhPanel = new JPanel();
      rhPanel.setBackground(bgColor);
      rhPanel.setLayout(new GridLayout(6, 1, 5, 5));

      JPanel r1Panel = new JPanel();
      JPanel r2Panel = new JPanel();
      JPanel r3Panel = new JPanel();
      JPanel r4Panel = new JPanel();
      JPanel r5Panel = new JPanel();
      JPanel r6Panel = new JPanel();
      r1Panel.setBackground(bgColor);
      r2Panel.setBackground(bgColor);
      r3Panel.setBackground(bgColor);
      r4Panel.setBackground(bgColor);
      r5Panel.setBackground(bgColor);
      r6Panel.setBackground(bgColor);

      JLabel lab12 = new JLabel();
      lab12.setFont(labFont);
      lab12.setText("Horiz. Datum  ");
      lab12.setToolTipText("Horizontal Datum");
      r1Panel.add(lab12);

      hdatCB = new JComboBox(hdatS);
      hdatCB.setMaximumRowCount(2);          //*** only 2 entries
      hdatCB.addItemListener(
         new ItemListener() {
            public void itemStateChanged(ItemEvent e1) {
               horIdx=hdatCB.getSelectedIndex();
            }
         }
      );
      r1Panel.add(hdatCB);
      rhPanel.add(r1Panel);
//*****
//****JLabel lab22 = new JLabel();           //*** used for error messages
             lab22 = new JLabel();
      lab22.setFont(mesFont);
      lab22.setText("Initializing");
      r2Panel.add(lab22);

      rhPanel.add(r2Panel);
//*****
      JLabel lab32 = new JLabel();
      lab32.setFont(labFont);
      lab32.setText("Input V-Datum ");
      lab32.setToolTipText("Input Vertical Datum");
      r3Panel.add(lab32);

      vindatCB = new JComboBox(vdatS);
      vindatCB.setMaximumRowCount(5);
      vindatCB.addItemListener(
         new ItemListener() {
            public void itemStateChanged(ItemEvent e2) {
               vinIdx=vindatCB.getSelectedIndex();
            }
         }
      );
      r3Panel.add(vindatCB);
      rhPanel.add(r3Panel);
//*****
      JLabel lab42 = new JLabel();
      lab42.setFont(labFont);
      lab42.setText("Output V-Datum");
      lab42.setToolTipText("Output Vertical Datum");
      r4Panel.add(lab42);

      voutdatCB = new JComboBox(vdatS);
      voutdatCB.setMaximumRowCount(5);
      voutdatCB.addItemListener(
         new ItemListener() {
            public void itemStateChanged(ItemEvent e3) {
               voutIdx=voutdatCB.getSelectedIndex();
            }
         }
      );
      r4Panel.add(voutdatCB);
      rhPanel.add(r4Panel);
//*****
      meterRButtn = new JRadioButton("Meters   ", true);
      meterRButtn.setBackground(bgColor);
      meterRButtn.setToolTipText("Units of Height");
      r5Panel.add(meterRButtn);

      feetRButtn  = new JRadioButton("Feet            ",   false);
      feetRButtn.setBackground(bgColor);
      feetRButtn.setToolTipText("Units of Height");
      r5Panel.add(feetRButtn);

      RadioButtonHandler handleRB = new RadioButtonHandler();
      meterRButtn.addItemListener(handleRB);
      feetRButtn.addItemListener(handleRB);

      fmRBGroup = new ButtonGroup();
      fmRBGroup.add(meterRButtn);
      fmRBGroup.add(feetRButtn);

      rhPanel.add(r5Panel);
//*****
      htRButtn = new JRadioButton("Height   ", true);
      htRButtn.setBackground(bgColor);
      htRButtn.setToolTipText("Heights are positive up");
      r6Panel.add(htRButtn);

      soundRButtn  = new JRadioButton("Soundings", false);
      soundRButtn.setBackground(bgColor);
      soundRButtn.setToolTipText("Soundings are positive down");
      r6Panel.add(soundRButtn);

      htRButtn.addItemListener(handleRB);
      soundRButtn.addItemListener(handleRB);

      hsRBGroup = new ButtonGroup();
      hsRBGroup.add(htRButtn);
      hsRBGroup.add(soundRButtn);

      rhPanel.add(r6Panel);

//*** build the card layout holder (left hand side)

      lhPanel.add("interact",lh0Panel);
      lhPanel.add("batch",   lh1Panel);

//*** place outer panels

      c.add(lhPanel);
      c.add(rhPanel);

      myCL = (CardLayout)lhPanel.getLayout();
      myCL.show(lhPanel,"interact");          //*** starts in interactive mode
      setSize(550, 310);                      //*** screen size
      show();

//*** initialize -- signal when finished

      initialize();
      lab22.setText("            ");    //*** clear earlier init. message
      repaint();
   }

   public void leftPanel0(JPanel lhPanel) {

//*** left panel -- interactive version

      lhPanel.setBackground(bgColor);
      lhPanel.setLayout(new GridLayout(6, 1, 5, 5));

      JPanel l1Panel = new JPanel();
      JPanel l2Panel = new JPanel();
      JPanel l3Panel = new JPanel();
      JPanel l4Panel = new JPanel();
      JPanel l5Panel = new JPanel();
      JPanel l6Panel = new JPanel();
      l1Panel.setBackground(bgColor);
      l2Panel.setBackground(bgColor);
      l3Panel.setBackground(bgColor);
      l4Panel.setBackground(bgColor);
      l5Panel.setBackground(bgColor);
      l6Panel.setBackground(bgColor);

      JLabel lab11 = new JLabel();
      lab11.setFont(labFont);
      lab11.setText("Latitude       ");
      lab11.setToolTipText("Decimal Degrees -or- " +
                           "Degrees, Decimal Minutes -or- " +
                           "Degrees, Minutes, Decimal Seconds");
      l1Panel.add(lab11);

      latTF = new JTextField(12);
      l1Panel.add(latTF);
      lhPanel.add(l1Panel);
//*****
      JLabel lab21 = new JLabel();
      lab21.setFont(labFont);
      lab21.setText("West Longitude ");
      lab21.setToolTipText("Positive West");
      l2Panel.add(lab21);

      lonTF = new JTextField(12);
      l2Panel.add(lonTF);
      lhPanel.add(l2Panel);
//*****
      JLabel lab31 = new JLabel();
      lab31.setFont(labFont);
      lab31.setText("Input Height   ");
      lab31.setToolTipText("meters or feet");
      l3Panel.add(lab31);

      hinTF = new JTextField(12);
      l3Panel.add(hinTF);
      lhPanel.add(l3Panel);
//*****
      JLabel lab41 = new JLabel();
      lab41.setFont(labFont);
      lab41.setText("Output Height  ");
      lab41.setToolTipText("meters or feet");
      l4Panel.add(lab41);

      houtTF = new JTextField(12);
      houtTF.setEditable(false);                //*** may not edit output
      l4Panel.add(houtTF);
      lhPanel.add(l4Panel);
//*****
      JLabel lab51 = new JLabel();
      lab51.setFont(labFont);
      lab51.setText(" ");                    //*** placeholder
      lhPanel.add(l5Panel);
//*****
      intrButtn = new JButton("Convert Vertical Datum");
      intrButtn.setToolTipText("Interactive Mode");
      lhPanel.add(intrButtn);

//*** button press handler

      ButtonHandler handler = new ButtonHandler();
      intrButtn.addActionListener(handler);
   }

   public void leftPanel1(JPanel lhPanel) {

//*** left panel -- batch version

      ButtonHandler handler = new ButtonHandler();

      lhPanel.setBackground(bgColor);
      lhPanel.setLayout(new GridLayout(6, 1, 5, 5));

      JPanel l1Panel = new JPanel();
      JPanel l2Panel = new JPanel();
      JPanel l3Panel = new JPanel();
      JPanel l4Panel = new JPanel();
      JPanel l5Panel = new JPanel();
      JPanel l6Panel = new JPanel();
      l1Panel.setBackground(bgColor);
      l2Panel.setBackground(bgColor);
      l3Panel.setBackground(bgColor);
      l4Panel.setBackground(bgColor);
      l5Panel.setBackground(bgColor);
      l6Panel.setBackground(bgColor);

      JLabel lab11 = new JLabel();
      lab11.setFont(labFont);
      lab11.setText("Input File            ");
      lab11.setToolTipText("Click button to find file");
      l1Panel.add(lab11);

      finButtn = new JButton("Browse");
      finButtn.setToolTipText("Browse Input File");
      l1Panel.add(finButtn);
      lhPanel.add(l1Panel);
      finButtn.addActionListener(handler);
//*****

      JLabel lab21 = new JLabel();
      lab21.setFont(labFont);
      lab21.setText("Output File           ");
      lab21.setToolTipText("Click button to find file");
      l2Panel.add(lab21);

      foutButtn = new JButton("Browse");
      foutButtn.setToolTipText("Browse Output File");
      l2Panel.add(foutButtn);
      lhPanel.add(l2Panel);
      foutButtn.addActionListener(handler);
//*****

      JLabel lab31 = new JLabel();
      lab31.setFont(labFont);
      lab31.setText("Input Filename ");
      lab31.setToolTipText("ASCII, 4-column text");
      l3Panel.add(lab31);

      finTF = new JTextField(12);
      l3Panel.add(finTF);
      lhPanel.add(l3Panel);
//*****
      JLabel lab41 = new JLabel();
      lab41.setFont(labFont);
      lab41.setText("Output Filename");
      lab41.setToolTipText("ASCII, 4-column text");
      l4Panel.add(lab41);

      foutTF = new JTextField(12);
      l4Panel.add(foutTF);
      lhPanel.add(l4Panel);
//*****
      latLonRButtn = new JRadioButton("Key,Lat,Lon,H  ", true);  //*** default
      latLonRButtn.setBackground(bgColor);
      latLonRButtn.setToolTipText("4-column format: Latitude before Longitude");
      l5Panel.add(latLonRButtn);

      lonLatRButtn  = new JRadioButton("Key,Lon,Lat,H  ",   false);
      lonLatRButtn.setBackground(bgColor);
      lonLatRButtn.setToolTipText("4-column format: Longitude before Latitude");
      l5Panel.add(lonLatRButtn);

      RadioButtonHandler handleRB = new RadioButtonHandler();
      latLonRButtn.addItemListener(handleRB);
      lonLatRButtn.addItemListener(handleRB);

      yxRBGroup = new ButtonGroup();
      yxRBGroup.add(latLonRButtn);
      yxRBGroup.add(lonLatRButtn);

      lhPanel.add(l5Panel);
//*****
      batButtn = new JButton("Batch File Conversion");
      batButtn.setToolTipText("Batch Mode");
      lhPanel.add(batButtn);
      batButtn.addActionListener(handler);
   }

   public void initialize() {

//*** initialize lots of stuff *****************************

      double glamn,glomn,dgla,dglo;
      int    nla,nlo;
      String myString=null;
      StringTokenizer myST;

//*** transformation parameters to NAD 83 (86)
//*** store: dx, dy, dz, rx, ry, rz, sk

      frameTS = new TSet[vdatS.length-NADIDX];   //*** transformations[9-->26]

//*** NAD 83 (86)
      frameTS[ 0] = new TSet(0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0);

//*** WGS_84(G1150) (use ITRF00)
      frameTS[ 1] = new TSet(0.9956, -1.9013, -0.52145,
                             1.256374e-7, 4.5701e-8, 5.6235e-8, 0.61504e-9);

//*** WGS_84(G873) (use ITRF94)
      frameTS[ 2] = new TSet(0.9910, -1.9072, -0.5129,
                             1.25033e-7, 4.6785e-8, 5.6529e-8, 0.0);

//*** WGS 84(G730) (use ITRF92)
      frameTS[ 3] = new TSet(0.9830, -1.9092, -0.5049,
                             1.25033e-7, 4.6785e-8, 5.6529e-8, 8.0e-10);

//*** WGS_84(orig) (use NAD 83)
      frameTS[ 4] = new TSet(0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0);

//*** WGS 72
      frameTS[ 5] = new TSet(0.0, 0.0, 4.5, 0.0, 0.0, -2.685868e-6, 2.263e-7);

//*** ITRF00
      frameTS[ 6] = new TSet(0.9956, -1.9013, -0.52145,
                             1.256374e-7, 4.5701e-8, 5.6235e-8, 0.61504e-9);

//*** ITRF97
      frameTS[ 7] = new TSet(0.9889, -1.9074, -0.50295,
                             1.256374e-7, 4.5701e-8, 5.6235e-8, -0.93496e-9);

//*** ITRF96
      frameTS[ 8] = new TSet(0.9910, -1.9072, -0.5129,
                             1.25033e-7, 4.6785e-8, 5.6529e-8, 0.0);

//*** ITRF94   (ITRF94 = ITRF96)
      frameTS[ 9] = new TSet(0.9910, -1.9072, -0.5129,
                             1.25033e-7, 4.6785e-8, 5.6529e-8, 0.0);

//*** ITRF93
      frameTS[10] = new TSet(0.9850, -1.9022, -0.4979,
                             1.23142e-7, 5.06635e-8, 5.18748e-8, -4.0e-10);

//*** ITRF92
      frameTS[11] = new TSet(0.9830, -1.9092, -0.5049,
                             1.25033e-7, 4.6785e-8, 5.6529e-8, 8.0e-10);

//*** ITRF91
      frameTS[12] = new TSet(0.9710, -1.9232, -0.4989,
                             1.25033e-7, 4.6785e-8, 5.6529e-8, -6.0e-10);

//*** ITRF90
      frameTS[13] = new TSet(0.9730, -1.9192, -0.4829,
                             1.25033e-7, 4.6785e-8, 5.6529e-8, -9.0e-10);

//*** ITRF89
      frameTS[14] = new TSet(0.9680, -1.9432, -0.4449,
                             1.25033e-7, 4.6785e-8, 5.6529e-8, -4.3e-9);

//*** ITRF88
      frameTS[15] = new TSet(0.9730, -1.9072, -0.4209, 
                             1.2552e-7, 4.6785e-8, 5.6529e-8, -7.4e-9);

//*** SIO/MIT 92  (use ITRF91)
      frameTS[16] = new TSet(0.9710, -1.9232, -0.4989,
                             1.25033e-7, 4.6785e-8, 5.6529e-8, -6.0e-10);

//*** NEOS 90  (use ITRF90)
      frameTS[17] = new TSet(0.9730, -1.9192, -0.4829,
                             1.25033e-7, 4.6785e-8, 5.6529e-8, -9.0e-10);

//*** PNEOS 90  (use ITRF90)
      frameTS[18] = new TSet(0.9730, -1.9192, -0.4829,
                             1.25033e-7, 4.6785e-8, 5.6529e-8, -9.0e-10);

//*** open grid files and read

      try {nclaBR = new BufferedReader(
                   new InputStreamReader(new FileInputStream ("ncla.gtx")));}
      catch (IOException e) {Dump.Fatal("File ncla.gtx could not be opened.");}

      try {ncloBR = new BufferedReader(
                   new InputStreamReader(new FileInputStream ("nclo.gtx")));}
      catch (IOException e) {Dump.Fatal("File nclo.gtx could not be opened.");}

      try {g99BR = new BufferedReader(
                    new InputStreamReader(new FileInputStream ("g99.gtx")));}
      catch (IOException e) {Dump.Fatal("File g99.gtx could not be opened.");}

      try {vcnBR = new BufferedReader(
                   new InputStreamReader(new FileInputStream ("vcn.gtx")));}
      catch (IOException e) {Dump.Fatal("File vcn.gtx could not be opened.");}

      try {sstBR = new BufferedReader(
                   new InputStreamReader(new FileInputStream ("sst.gtx")));}
      catch (IOException e) {Dump.Fatal("File sst.gtx could not be opened.");}

      try {mllwBR = new BufferedReader(
                    new InputStreamReader(new FileInputStream ("mllw.gtx")));}
      catch (IOException e) {Dump.Fatal("File mllw.gtx could not be opened.");}

      try {mlwBR  = new BufferedReader(
                    new InputStreamReader(new FileInputStream ("mlw.gtx")));}
      catch (IOException e) {Dump.Fatal("File mlw.gtx could not be opened.");}

      try {mhwBR  = new BufferedReader(
                    new InputStreamReader(new FileInputStream ("mhw.gtx")));}
      catch (IOException e) {Dump.Fatal("File mhw.gtx could not be opened.");}

      try {mhhwBR = new BufferedReader(
                    new InputStreamReader(new FileInputStream ("mhhw.gtx")));}
      catch (IOException e) {Dump.Fatal("File mhhw.gtx could not be opened.");}

      try {mtlBR  = new BufferedReader(
                    new InputStreamReader(new FileInputStream ("mtl.gtx")));}
      catch (IOException e) {Dump.Fatal("File mtl.gtx could not be opened.");}

      try {dtlBR  = new BufferedReader(
                    new InputStreamReader(new FileInputStream ("dtl.gtx")));}
      catch (IOException e) {Dump.Fatal("File dtl.gtx could not be opened.");}

      nclaG=Greader("ncla.gtx", nclaBR);   //*** read nadcon, latitude
      ncloG=Greader("nclo.gtx", ncloBR);   //*** read nadcon, longitude
      g99G =Greader("g99.gtx",  g99BR);    //*** read geoid
      vcnG =Greader("vcn.gtx",  vcnBR);    //*** read vertcon
      sstG =Greader("sst.gtx",  sstBR);    //*** read sea surf topo
      mllwG=Greader("mllw.gtx", mllwBR);   //*** read mean lower low water
      mlwG =Greader("mlw.gtx",  mlwBR);    //*** read mean low water
      mhwG =Greader("mhw.gtx",  mhwBR);    //*** read mean high water
      mhhwG=Greader("mhhw.gtx", mhhwBR);   //*** read mean higher high water
      mtlG =Greader("mtl.gtx",  mtlBR);    //*** read mean tide level
      dtlG =Greader("dtl.gtx",  dtlBR);    //*** read diurnal tide level

      nclaBR=null;                         //*** close files
      ncloBR=null;
      g99BR =null;
      vcnBR =null;
      sstBR =null;
      mllwBR=null;
      mlwBR =null;
      mhwBR =null;
      mhhwBR=null;
      mtlBR =null;
      dtlBR =null;

   } //***end initialize

   private Grid Greader(String fname, BufferedReader myBR) {

//*** basic housekeeping to read a grid

      double glamn=0.0,glomn=0.0,dgla=0.0,dglo=0.0;
      int    nla=0,nlo=0;
      String myString=null;

      try {myString=myBR.readLine();}
      catch (IOException e) {Dump.Fatal(fname + " header input error.");}

      StringTokenizer myST= new StringTokenizer( myString.trim() );
      if(myST.countTokens() != 6) Dump.Fatal(fname + " header input error.");

      try{ glamn =Double.parseDouble(myST.nextToken()); }
      catch(NumberFormatException nfe) {Dump.Fatal(fname + " header format error.");}
//*** input pos. EAST
      try{ glomn =Double.parseDouble(myST.nextToken()); }
      catch(NumberFormatException nfe) {Dump.Fatal(fname + " header format error.");}

      try{ dgla  =Double.parseDouble(myST.nextToken()); }
      catch(NumberFormatException nfe) {Dump.Fatal(fname + " header format error.");}

      try{ dglo  =Double.parseDouble(myST.nextToken()); }
      catch(NumberFormatException nfe) {Dump.Fatal(fname + " header format error.");}

      try{ nla   =Integer.parseInt(myST.nextToken()); }
      catch(NumberFormatException nfe) {Dump.Fatal(fname + " header format error.");}

      try{ nlo   =Integer.parseInt(myST.nextToken()); }
      catch(NumberFormatException nfe) {Dump.Fatal(fname + " header format error.");}

      Grid myG = new Grid(glamn, glomn, dgla, dglo, nla, nlo);

      try {myG.readGrid(myBR);}
      catch (IOException e)             {Dump.Fatal(fname + " grid input error.");}
      catch (NumberFormatException nfe) {Dump.Fatal(fname + " grid format error.");}

      return myG;
   }

//***************************************************************************
//*** inner classes *********************************************************
//***************************************************************************
   private class ButtonHandler implements ActionListener {

//*** button event handling

      public void actionPerformed(ActionEvent e) {

         double glad=0.0,glod=0.0,eht=0.0,glodw=0.0,htout,elin=0.0,elout;
         String glaS,gloS,ehtS,myStr=null, elinS, eloutS;
         String keyS;                 //*** GIS data base support (batch only)
         boolean noProb;

//***********************************
//**** interactive mode *************
//***********************************
         if      (e.getSource() ==  intrButtn){
            lab22.setText("                   ");       //*** clear

//*** latitude box

            glaS=latTF.getText();
            if(glaS.equals("")){
               glaS="0.0";
               latTF.setText(glaS);
            }
            glad=parseDMS(glaS);
            if(glad < -90.0) glad = DNULL;          //*** range checks
            if(glad >  90.0) glad = DNULL;
            if(glad <= DNULL) lab22.setText("Lat. Format Error  ");

//*** longitude box
         
            gloS=lonTF.getText();
            if(gloS.equals("")){
               gloS="0.0";
               lonTF.setText(gloS);
            }
            glodw=parseDMS(gloS);
            if(glodw < -360.0) glodw = DNULL;       //*** range checks
            if(glodw >  360.0) glodw = DNULL;
            if(glodw <= DNULL) lab22.setText("Long. Format Error ");
            glod =360.0 - glodw;                   //*** pos WEST to EAST
            if(glod  >= 360.0) glod  = glod-360.0;

//*** input height box

            ehtS=hinTF.getText();
            if(ehtS.equals("")){
               ehtS="0.0";
               hinTF.setText(ehtS);
            }
            try{eht=Double.parseDouble(ehtS);}
            catch(NumberFormatException nfe){
               eht = DNULL;
               lab22.setText("Ht. Format Error   ");
            }

//*** only convert if no prior errors
//*** in/out elevations are meters at this point

            if(glad <= DNULL || glodw <= DNULL || eht <= DNULL){
               htout = DNULL;
            }else{
               if(iFeet != 0) eht=eht/TOFEET;        //*** input feet to meters
               GPos inGP = new GPos(glad, glod, eht);
               htout = inGP.convert(frameTS, nclaG, ncloG, g99G, vcnG, sstG, mllwG, 
                                    mlwG, mhwG, mhhwG, mtlG, dtlG,
                                    horIdx, vinIdx, voutIdx, iSound);
               if(htout <= Grid.GNULL) lab22.setText("Grid Boundary Error");
            }
            if(htout > Grid.GNULL){
               if(iFeet != 0) htout=htout*TOFEET;  //*** meters to output feet
            }
            houtTF.setText(dec4DF.format(htout));  //*** output field (4 dec)

//***********************************
//**** batch mode *******************
//***********************************
         }else if(e.getSource() ==  batButtn) {
            try {finBR = new BufferedReader(new InputStreamReader(
                new FileInputStream (finTF.getText())));}
            catch (IOException ex) {
                Dump.Fatal("Input batch file could not be opened.");}

            try {foutBR = new BufferedWriter(new OutputStreamWriter(
                new FileOutputStream (foutTF.getText())));}
            catch (IOException ex) {
                Dump.Fatal("Output batch file could not be opened.");}

//*** infinite loop until EOF

            int nrec =0;                              //*** count the records
            int nprob=0;                              //*** count the problems
infLoop:{while(true){
            noProb=true;
            try {myStr=finBR.readLine();}
            catch (IOException ex) {
               Dump.Fatal("Batch file input format error 1.");
            }
            if(myStr==null){                            //*** EOF encountered
               closeFiles(finBR);
               break infLoop;
            }

//*** specify delimiter set

            StringTokenizer myST= new StringTokenizer( myStr.trim(), 
                                                       " \t\n\r\f,;:" );
            if(myST.countTokens() != 4) 
               Dump.Fatal("Batch file input format error 2.");

            keyS=myST.nextToken();                //*** GIS data base support

            if(iLonLat == 0){                     //*** key,lat,lon,h
               glaS=myST.nextToken();
               try{glad=Double.parseDouble(glaS);}
               catch(NumberFormatException nfe) {noProb=false;}

               gloS=myST.nextToken();              //*** pos WEST -or- lead minus
               try{glodw=Double.parseDouble(gloS);}
               catch(NumberFormatException nfe) {noProb=false;}
               if(noProb){
                  if(glodw >= -180.0 && glodw < 0.0) glodw=-glodw;  //*** lead minus
                  glod =360.0 - glodw;                        //*** pos WEST to EAST
               }

            }else{                                 //*** key,lon,lat,h
               gloS=myST.nextToken();              //*** pos WEST -or- lead minus
               try{glodw=Double.parseDouble(gloS);}
               catch(NumberFormatException nfe) {noProb=false;}
               if(noProb){
                  if(glodw >= -180.0 && glodw < 0.0) glodw=-glodw;  //*** lead minus
                  glod =360.0 - glodw;                        //*** pos WEST to EAST
               }

               glaS=myST.nextToken();
               try{glad=Double.parseDouble(glaS);}
               catch(NumberFormatException nfe) {noProb=false;}
            }

            elinS=myST.nextToken();
            try{elin=Double.parseDouble(elinS);}
            catch(NumberFormatException nfe) {noProb=false;}

//*** only convert if no prior errors         (otherwise set to null)
//*** in/out elevations are meters at this point

            nrec=nrec+1;
            if(noProb){
               if(iFeet != 0) elin=elin/TOFEET;      //*** input feet to meters
               GPos inGP = new GPos(glad, glod, elin);
               elout = inGP.convert(frameTS, nclaG, ncloG, g99G, vcnG, sstG, mllwG, 
                                    mlwG, mhwG, mhhwG, mtlG, dtlG,
                                    horIdx, vinIdx, voutIdx, iSound);
               if(elout <= Grid.GNULL) elout = DNULL;
            }else{
               elout = DNULL;
            }
            if(elout != DNULL){
               if(iFeet  != 0) elout=elout*TOFEET;  //*** meters to output feet
            }else{
               nprob=nprob+1;                      //*** count the problems
            }
            eloutS=dec4DF.format(elout);           //*** output field (4 dec)

//*** note to me -- could be more elegent in generating 4 column text (debug)

            myStr=myStr.substring(0, myStr.lastIndexOf(elinS))+eloutS;

//*** output line

            try {foutBR.write(myStr, 0, myStr.length());}
            catch (IOException ex) {
               Dump.Fatal("Batch file output error 1.");
            }
            try {foutBR.newLine();}
            catch (IOException ex) {
               Dump.Fatal("Batch file newLine output error");
            }
         }}//*** end while loop

///*** message box -- done processing

         JOptionPane.showMessageDialog(VDatum.this,
            "Processed " + nrec + " records\n"+

            (iSound != 0 ? "Sounding" : "Height") +
            " data, units of "+
            (iFeet != 0 ? "Feet" : "Meters") + "\n"

            + nprob + " had problems",

            "Batch Processing Report", JOptionPane.PLAIN_MESSAGE);

         lab22.setText("                   ");
         repaint();

//***********************************
//**** input file browse ************
//***********************************
         }else if(e.getSource() ==  finButtn) {
            inFile=myOpenFile();
            if(inFile != null && inFile.isFile()) 
                  finTF.setText(inFile.getAbsolutePath());

//***********************************
//**** output file browse ***********
//***********************************
         }else if(e.getSource() ==  foutButtn){
            outFile=mySaveFile();
            if(outFile != null && outFile.isFile())
                  foutTF.setText(outFile.getAbsolutePath());

//***********************************
//**** program error ****************
//***********************************
         }else{
            Dump.Prog("mode error: ButtonHandler"); //***should never get here
         }
      }//*** end actionPerformed

      private void closeFiles(BufferedReader finBR){
         try {
            finBR.close();
            foutBR.flush();
            foutBR.close();
         }
         catch (IOException e) {Dump.Fatal("Error closing files");}
      }

      public double parseDMS(String myString) {

//*** parse a DD.dddd, DD MM.mmmm, or DD MM SS.ssss

         int     ideg=0,imin=0;
         double  deg=0.0,min=0.0,sec=0.0;
         boolean isNeg=false;               //*** assume no negative sign

         StringTokenizer myST= new StringTokenizer( myString.trim() );
         int ntok = myST.countTokens();

//*** decimal degrees

         if(ntok == 1){
            try{ deg=Double.parseDouble(myST.nextToken()); }
            catch(NumberFormatException nfe) {return DNULL;}
            return deg;                                //*** need this

//*** degrees, decimal minutes  (use sign of degrees)

         }else if(ntok == 2){
            try{ ideg=Integer.parseInt(myST.nextToken()); }
            catch(NumberFormatException nfe) {return DNULL;}
            if(ideg < 0.0){
               isNeg= true;
               ideg = Math.abs(ideg);
            }
            try{ min=Double.parseDouble(myST.nextToken()); }
            catch(NumberFormatException nfe) {return DNULL;}
            min = Math.abs(min);
            deg = ideg + min/60.0;

//*** degrees, minutes, decimal seconds  (use sign of degrees)

         }else if(ntok == 3){
            try{ ideg=Integer.parseInt(myST.nextToken()); }
            catch(NumberFormatException nfe) {return DNULL;}
            if(ideg < 0.0){
               isNeg= true;
               ideg = Math.abs(ideg);
            }
            try{ imin=Integer.parseInt(myST.nextToken()); }
            catch(NumberFormatException nfe) {return DNULL;}
            imin = Math.abs(imin);
            try{ sec=Double.parseDouble(myST.nextToken()); }
            catch(NumberFormatException nfe) {return DNULL;}
            sec = Math.abs(sec);
            deg = ideg + imin/60.0 + sec/3600.0;

//*** invalid number of tokens -- error return

         }else{
            return DNULL;
         }

         if(isNeg){
            return -deg;         //*** return with sign of the degrees
         }else{
            return deg;
         }
      }

      private File myOpenFile() {

//*** display file chooser dialog box

         JFileChooser fileChooser = new JFileChooser();
         fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
         int result = fileChooser.showOpenDialog(null);

//*** user clicked Cancel button on dialog

         if (result == JFileChooser.CANCEL_OPTION)
            return null;
         else
            return fileChooser.getSelectedFile();
      }

      private File mySaveFile() {

//*** display file chooser dialog box

         JFileChooser fileChooser = new JFileChooser();
         fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
         int result = fileChooser.showSaveDialog(null);

//*** user clicked Save button on dialog

         if (result == JFileChooser.APPROVE_OPTION){

            if(fileChooser.getSelectedFile().exists()){
               return fileChooser.getSelectedFile();         //*** to old file
            }else{
               try{
                  if(fileChooser.getSelectedFile().createNewFile()){
                     return fileChooser.getSelectedFile();   //*** to new file
                  }else{
                     return null;                            //*** fail
                  }
               }
               catch (IOException e) 
                      {Dump.Fatal("New output file could not be created.");}
            }
         }//*** endif approve

         return null;   //*** did not approve a file name
      }

   }//*** end ButtonHandler
//***************************************************************************
   private class RadioButtonHandler implements ItemListener {

//*** radio button event handling

      public void itemStateChanged(ItemEvent e) {
         if(e.getSource() == meterRButtn)
            iFeet=0;
         else if(e.getSource() == feetRButtn)
            iFeet=1;
         else if(e.getSource() == htRButtn)
            iSound=0;
         else if(e.getSource() == soundRButtn)
            iSound=1;
         else if(e.getSource() == latLonRButtn)
            iLonLat=0;
         else if(e.getSource() == lonLatRButtn)
            iLonLat=1;
         else
            Dump.Prog("mode error: RadioButtonHandler");  //***should never get here

         lab22.setText("                   ");       //*** clear any errors
         repaint();
      }
   }//*** end RadioButtonHandler
//***************************************************************************
   class ItemHandler implements ActionListener {
      public void actionPerformed(ActionEvent e) {

//*** alter GUI based on mode selected from menu

         CardLayout myCL = (CardLayout)lhPanel.getLayout();

         if      (modeItems[0].isSelected()) {
            iMode=0;
            myCL.show(lhPanel,"interact");
         }else if(modeItems[1].isSelected()) {
            iMode=1;
            myCL.show(lhPanel,"batch");
         }else{
            Dump.Prog("mode error: ItemHandler");   //***should never get here
         }

         lab22.setText("                   ");  //*** clear any error messages
         repaint();
      }
   }//*** end ItemHandler
}//***endclass
//************************************************************************
class GPos {

//*** geodetic position class -- GPos

   private double glad,glod,eht;

//*** constructors

   public GPos(double gladx, double glodx, double ehtx) {
      glad=gladx;
      glod=glodx;
      eht =ehtx;
   }

//*** set function

   public void putGPos(double gladx, double glodx, double ehtx){
      glad=gladx;
      glod=glodx;
      eht =ehtx;
   }

//*** view functions

   public double getLat()  {return glad;}
   public double getLon()  {return glod;}
   public double getEht()  {return eht;}

//*********************
//*** convert functions
//*********************

   public double convert(TSet frameTS[], Grid nclaG, Grid ncloG, Grid g99G, 
                         Grid vcnG, Grid sstG, Grid mllwG, Grid mlwG, 
                         Grid mhwG, Grid mhhwG, Grid mtlG, Grid dtlG,
                         int horIdx, int vinIdx, int voutIdx, int iSound) {

//*** convert vertical to desired datum

      GPos wrkGP;
      double delt;  //*** height conversion (m)

      int iI=vinIdx, oI=voutIdx, ipI, opI;           //*** path indicies
      double gla=glad, glo=glod, ht = eht;           //*** working variables
      double htin = eht;                             //*** copy input height

//*** establish conversion path

      ipI = principal(iI);
      opI = principal(oI);

//*** phase 0 -- convert input lat/lon from NAD27 (if needed)
//***         -- units of arc seconds

      if(horIdx == 1){

         double gladx,glodx;            //*** temp data holders for NAD 83

         double deltla=nclaG.bqterpG(glad, glod);  //*** units of arcsec
         if(deltla > Grid.GNULL){
            gladx=glad+deltla/3600.0;         //*** add to 27 to get 83
         }else{         
            return Grid.GNULL;  
         }

         double deltlo=ncloG.bqterpG(glad, glod);  //*** units of arcsec
         if(deltlo > Grid.GNULL){
            glodx=glod-deltlo/3600.0;         //*** minus, NADCON was pos W.
         }else{         
            return Grid.GNULL;  
         }
         glad=gladx;           //*** glad/glod used for all grid lookups
         glod=glodx;
         gla =gladx;           //*** gla/glo used for 3-D transformations
         glo =glodx;
      }

      wrkGP = new GPos(gla, glo, ht);

//*** phase 1 -- convert input height to input principal height

      if(iI != ipI){
         if(orthoG(ipI)){
            delt=vcnG.bqterpG(glad, glod);        //*** units of mm
            if(delt > Grid.GNULL){
               ht=ht+delt/1000.0;                  //*** add to 29 to get 88
            }else{         
               return Grid.GNULL;  
            }
         }else if(tidalG(ipI)){

//*** MLLW to LMSL
//*** 23-aug-00  fixed sign error on application of tide datums

            if(iI == VDatum.TIDIDX){
               delt=mllwG.blterpG(glad, glod);
               if(delt > Grid.GNULL){
                  ht=ht+delt;             //*** add to mllw to get lmsl
               }else{         
                  return Grid.GNULL;  
               }
            }//*** endif on iI

//*** MLW to LMSL

            if(iI == VDatum.TIDIDX+1){
               delt=mlwG.blterpG(glad, glod);
               if(delt > Grid.GNULL){
                  ht=ht+delt;             //*** add to mlw to get lmsl
               }else{         
                  return Grid.GNULL;  
               }
            }//*** endif on iI

//*** MTL to LMSL

            if(iI == VDatum.TIDIDX+3){
               delt=mtlG.blterpG(glad, glod);
               if(delt > Grid.GNULL){
                  ht=ht+delt;             //*** add to mtl to get lmsl
               }else{         
                  return Grid.GNULL;  
               }
            }//*** endif on iI

//*** DTL to LMSL

            if(iI == VDatum.TIDIDX+4){
               delt=dtlG.blterpG(glad, glod);
               if(delt > Grid.GNULL){
                  ht=ht+delt;             //*** add to dtl to get lmsl
               }else{         
                  return Grid.GNULL;  
               }
            }//*** endif on iI

//*** MHW to LMSL

            if(iI == VDatum.TIDIDX+5){
               delt=mhwG.blterpG(glad, glod);
               if(delt > Grid.GNULL){
                  ht=ht+delt;             //*** add to mhw to get lmsl
               }else{         
                  return Grid.GNULL;  
               }
            }//*** endif on iI

//*** MHHW to LMSL

            if(iI == VDatum.TIDIDX+6){
               delt=mhhwG.blterpG(glad, glod);
               if(delt > Grid.GNULL){
                  ht=ht+delt;             //*** add to mhhw to get lmsl
               }else{         
                  return Grid.GNULL;  
               }
            }//*** endif on iI

//*** 3-D conversions

         }else if(ellipG(ipI)){
            wrkGP = wrkGP.newframe(frameTS, iI, ipI, true);  //*** to NAD83
            gla=wrkGP.getLat();
            glo=wrkGP.getLon();
            ht =wrkGP.getEht();
         }else{
            Dump.Prog("group error 1: ipI= " +ipI);  //***should never get here
         }
      }

//*** phase 2 -- convert input principal height to output principal height

      if(ipI != opI){
         if      (ipI == VDatum.NADIDX){                //*** from NAD 83
            delt=g99G.bqterpG(glad, glod);
            if(delt > Grid.GNULL){
               ht=ht-delt;                  //*** subtract from 83 to get 88
            }else{
               return Grid.GNULL;  
            }
            if(opI == VDatum.MSLIDX){
               delt=sstG.blterpG(glad, glod);
               if(delt > Grid.GNULL){
//*** 25-jan-02  fixed sign error in following line
                  ht=ht+delt;               //*** add to 88 to get LMSL
               }else{ 
                  return Grid.GNULL;  
               }
            }

         }else if(ipI ==  VDatum.MSLIDX){               //*** from LMSL
            delt=sstG.blterpG(glad, glod);
            if(delt > Grid.GNULL){
               ht=ht-delt;                  //*** subtract from LMSL to get 88
            }else{
               return Grid.GNULL;  
            }
            if(opI == VDatum.NADIDX){
               delt=g99G.bqterpG(glad, glod);
               if(delt > Grid.GNULL){
                  ht=ht+delt;               //*** add to 88 to get 83
               }else{          
                  return Grid.GNULL;  
               }
            }

         }else if(ipI ==  VDatum.ORTIDX){               //*** from NAVD 88
            if(opI == VDatum.MSLIDX){
               delt=sstG.blterpG(glad, glod);
               if(delt > Grid.GNULL){
                  ht=ht+delt;               //*** add to 88 to get LMSL
               }else{        
                  return Grid.GNULL;  
               }
            }else if(opI == VDatum.NADIDX){
               delt=g99G.bqterpG(glad, glod);
               if(delt > Grid.GNULL){
                  ht=ht+delt;               //*** add to 88 to get 83
               }else{          
                  return Grid.GNULL;  
               }
            }else{
               Dump.Prog("group error 2: opI= " +opI);  //***should never get here
            }

         }else{
            Dump.Prog("group error 4: ipI= " +ipI);  //***should never get here
         }//*** end multiple if set

         wrkGP = new GPos(gla, glo, ht);   //*** update height in work object
      }

//*** phase 3 -- convert output principal height to output height

      if(opI != oI){
         if(orthoG(opI)){
            delt=vcnG.bqterpG(glad, glod);        //*** units of mm
            if(delt > Grid.GNULL){
               ht=ht-delt/1000.0;          //*** subtract from 88 to get 29
            }else{
               return Grid.GNULL;  
            }
         }else if(tidalG(opI)){

//*** LMSL to MLLW
//*** 23-aug-00  fixed sign error on application of tide datums

            if(oI == VDatum.TIDIDX){
               delt=mllwG.blterpG(glad, glod);
               if(delt > Grid.GNULL){
                  ht=ht-delt;             //*** subtract from lmsl to get mllw
               }else{         
                  return Grid.GNULL;  
               }
            }//*** endif on oI

//*** LMSL to MLW

            if(oI == VDatum.TIDIDX+1){
               delt=mlwG.blterpG(glad, glod);
               if(delt > Grid.GNULL){
                  ht=ht-delt;             //*** subtract from lmsl to get mlw
               }else{         
                  return Grid.GNULL;  
               }
            }//*** endif on oI

//*** LMSL to MTL

            if(oI == VDatum.TIDIDX+3){
               delt=mtlG.blterpG(glad, glod);
               if(delt > Grid.GNULL){
                  ht=ht-delt;             //*** subtract from lmsl to get mtl
               }else{         
                  return Grid.GNULL;  
               }
            }//*** endif on oI

//*** LMSL to DTL

            if(oI == VDatum.TIDIDX+4){
               delt=dtlG.blterpG(glad, glod);
               if(delt > Grid.GNULL){
                  ht=ht-delt;             //*** subtract from lmsl to get dtl
               }else{         
                  return Grid.GNULL;  
               }
            }//*** endif on oI

//*** LMSL to MHW

            if(oI == VDatum.TIDIDX+5){
               delt=mhwG.blterpG(glad, glod);
               if(delt > Grid.GNULL){
                  ht=ht-delt;             //*** subtract from lmsl to get mhw
               }else{         
                  return Grid.GNULL;  
               }
            }//*** endif on oI

//*** LMSL to MHHW

            if(oI == VDatum.TIDIDX+6){
               delt=mhhwG.blterpG(glad, glod);
               if(delt > Grid.GNULL){
                  ht=ht-delt;             //*** subtract from lmsl to get mhhw
               }else{         
                  return Grid.GNULL;  
               }
            }//*** endif on oI

//*** 3-D conversions

         }else if(ellipG(opI)){
            wrkGP = wrkGP.newframe(frameTS, opI, oI, false);  //*** from NAD83
            gla=wrkGP.getLat();
            glo=wrkGP.getLon();
            ht =wrkGP.getEht();
         }else{
            Dump.Prog("group error 3: opI= " +opI);  //***should never get here
         }
      }

      if(iSound != 0) ht = htin - (ht - htin);                //*** soundings
      return ht;
   }

   public int principal(int i) {

//*** what is principal node for a conversion group?
//*** note: update if datum list is changed
      
//*** orthometric group -- return NAVD 88

      if(orthoG(i)){
        return VDatum.ORTIDX;

//*** tidal group -- return LMSL

      }else if(tidalG(i)){
        return VDatum.MSLIDX;

//*** ellipsoidal group -- return NAD 83

      }else if(ellipG(i)){
        return VDatum.NADIDX;
      }

//*** should never get here  (won't get to do return)

      Dump.Prog("illegal datum code " + i);
      return 0;
   }

   public boolean orthoG(int i) {

//*** is this an orthometric group?

      if(i >= VDatum.ORTIDX && i < VDatum.TIDIDX){return true;}
      return false;
   }

   public boolean tidalG(int i) {

//*** is this an tidal group?

      if(i >= VDatum.TIDIDX && i < VDatum.NADIDX){return true;}
      return false;
   }

   public boolean ellipG(int i) {

//*** is this an ellipsoidal group?

      if(i >= VDatum.NADIDX && i < VDatum.vdatS.length){return true;}
      return false;
   }

   public GPos newframe(TSet frameTS[], int inI, int outI, boolean toNad) {
 
//*** compute geodetic pos in new frame

      RPos myRP;

//*** convert to x,y,z

      myRP = this.toRP();

//*** frame translation   (boolean true --> to NAD 83)

      if(outI == VDatum.NADIDX){
         myRP = myRP.toNadRP(frameTS[inI -VDatum.NADIDX], true);
      }else if(inI == VDatum.NADIDX){
         myRP = myRP.toNadRP(frameTS[outI-VDatum.NADIDX], false);
      }else{               //***should never get here
         Dump.Prog("index error: inI,outI= " +inI +" " +outI);
      }

//*** convert back to geodetic

      return myRP.toGP();
   }

   public RPos toRP() {
 
//*** compute x,y,z
//*** ref p.17 geometric geodesy notes vol 1, osu, rapp
 
      double gla,glo,ht,sla,cla,w,n,x,y,z;

//*** constant variables for GRS-80 ellipsoid

      final double A=6378137.0, E2=0.006694380022903416, AE2=A*E2;

      final double RAD=180.0/Math.PI;                     //*** radians to deg.

      gla=glad/RAD;
      glo=glod/RAD;

      sla=Math.sin(gla);
      cla=Math.cos(gla);
      w  =Math.sqrt(1.0-E2*sla*sla);
      n  =A/w;
 
      x=(n          +eht)*cla*Math.cos(glo);
      y=(n          +eht)*cla*Math.sin(glo);
      z=(n*(1.0-E2) +eht)*sla;

      return new RPos(x,y,z);
   }
}//***endclass GPos
//************************************************************************
class RPos {

//*** rectangular (3-D) position class -- RPos

   private double x,y,z;

//*** constructor

   public RPos(double xin, double yin, double zin) {
      x=xin;
      y=yin;
      z=zin;
   }

//*** set functions

   public void putRPos(double xin, double yin, double zin) {
      x=xin;
      y=yin;
      z=zin;
   }

//*** view functions

   public double getX()  {return x;}
   public double getY()  {return y;}
   public double getZ()  {return z;}

//*** convert functions *******************************

   public RPos toNadRP(TSet myTS, boolean toNad){

//*** apply frame transformation  (boolean true --> to NAD 83)

      double dx,dy,dz,rx,ry,rz,ds;

      if(toNad){
         dx =       myTS.getTSdx();
         dy =       myTS.getTSdy();
         dz =       myTS.getTSdz();
         rx =       myTS.getTSrx();
         ry =       myTS.getTSry();
         rz =       myTS.getTSrz();
         ds = 1.0 + myTS.getTSsk();
      }else{
         dx =     - myTS.getTSdx();
         dy =     - myTS.getTSdy();
         dz =     - myTS.getTSdz();
         rx =     - myTS.getTSrx();
         ry =     - myTS.getTSry();
         rz =     - myTS.getTSrz();
         ds = 1.0 - myTS.getTSsk();
      }

      double xp = dx + ds*x + rz*y - ry*z;
      double yp = dy - rz*x + ds*y + rx*z;
      double zp = dz + ry*x - rx*y + ds*z;

      return new RPos(xp,yp,zp);
   }

   public GPos toGP() {
 
//*** convert x,y,z into geodetic lat, lon, and ellip. ht
//*** output decimal degrees -- long. pos EAST
 
//*** ref: eq a.4b, p. 132, appendix a, osu #370
//*** ref: geom geod notes gs 658, rapp

      final  double MAXINT=10,TOL=1.0e-13;
      double p,tgla,tglax,sla,w,n;
      double gla,glo,glad,glod,eht;
      int    icount=0;

//*** constant variables for GRS-80 ellipsoid

      final double A=6378137.0, E2=0.006694380022903416, AE2=A*E2;

      final double RAD=180.0/Math.PI;                     //*** radians to deg.
 
//*** compute initial estimate of reduced latitude  (eht=0)
 
      p=Math.sqrt(x*x+y*y);
      tgla=z/p/(1.0-E2);
 
//*** iterate to convergence, or to max # iterations

      while(icount < MAXINT){
         tglax=tgla;
         tgla=z/(p-(AE2/Math.sqrt(1.0+(1.0-E2)*tgla*tgla)));
         icount=icount+1;

//*** convergence test

         if(Math.abs(tgla-tglax) <= TOL){
            gla=Math.atan(tgla);
            sla=Math.sin(gla);
            glo=Math.atan2(y,x);
            w=Math.sqrt(1.0-E2*sla*sla);
            n=A/w;
            if(Math.abs(gla) < 0.7854){
               eht=p/Math.cos(gla)-n;
            }else{
               eht=z/sla-n+E2*n;
            }

            glad=gla*RAD;
            glod=glo*RAD;
            if(glod >= 360.0) glod=glod-360.0;
            if(glod <    0.0) glod=glod+360.0;
            return new GPos(glad, glod, eht);
         }
      }//***endwhile
 
//*** too many iterations -- probably too close to pole

      return new GPos(0.0, 0.0, 0.0);  //*** could reset to either pole
   }
}//***endclass RPos
//************************************************************************
class TSet {

//*** Frame Transformation class -- TSet

   private double dx,dy,dz,  //*** meters
                  rx,ry,rz,  //*** radians
                  sk;        //*** delta scale (not PPM)

//*** constructor

   public TSet(double dx0, double dy0, double dz0, 
               double rx0, double ry0, double rz0, double sk0) {
      dx=dx0;
      dy=dy0;
      dz=dz0;
      rx=rx0;
      ry=ry0;
      rz=rz0;
      sk=sk0;
   }

//*** set function

   public void putTSet(double dx0, double dy0, double dz0, 
               double rx0, double ry0, double rz0, double sk0) {
      dx=dx0;
      dy=dy0;
      dz=dz0;
      rx=rx0;
      ry=ry0;
      rz=rz0;
      sk=sk0;
   }

//*** view functions

   public double getTSdx()  {return dx;}
   public double getTSdy()  {return dy;}
   public double getTSdz()  {return dz;}
   public double getTSrx()  {return rx;}
   public double getTSry()  {return ry;}
   public double getTSrz()  {return rz;}
   public double getTSsk()  {return sk;}

}//***endclass GPos
//************************************************************************
class Grid {

//*** Gridded data class -- Grid
//***  uses fortran style index (0 row/column not used)

   private double glamn,glamx,dgla;   //*** decimal degrees
   private double glomn,glomx,dglo;   //*** decimal degrees, pos EAST
   private int    nla,nlo;            //*** # rows/columns
   private double g[][];              //*** ir (S/N), jc (W/E) indicies

   public static final double GNULL = -999999.0;  //*** null value
   public static final double NULL2 = -80.0;      //*** imbedded null value
   public static final double GTOL = 1.E-14;      //*** epsilon
   public static final double BTOL = 1.E-4;       //*** boundary tolerance

//*** constructor

   public Grid(double glamn0, double glomn0, double dgla0, double dglo0, 
               int    nla0,   int    nlo0) {

      glamn = glamn0;
      dgla  = dgla0;
      glomn = glomn0;
      dglo  = dglo0;
      nla   = nla0;
      nlo   = nlo0;
      glamx = glamn+(nla-1)*dgla+GTOL;
      glomx = glomn+(nlo-1)*dglo+GTOL;

      g     = new double[nla+1][nlo+1];       //*** fortran-style index
   }

//*** set functions

   public void putGHead(double glamn0, double glomn0, 
                        double dgla0,  double dglo0) {

      glamn = glamn0;
      dgla  = dgla0;
      glomn = glomn0;
      dglo  = dglo0;
      glamx = glamn+(nla-1)*dgla;    //*** nla,nlo already set by constructor
      glomx = glomn+(nlo-1)*dglo;
   }

   public void putGval(int ila, int ilo, double val) {

//*** fortran-style index
      if(ila < 1)   Dump.Prog("index error: ila=     " +ila);    
      if(ila > nla) Dump.Prog("index error: ila,nla= " +ila +" " +nla);
      if(ilo < 1)   Dump.Prog("index error: ilo=     " +ilo);
      if(ilo > nlo) Dump.Prog("index error: ilo,nlo= " +ilo +" " +nlo);

      g[ila][ilo] = val;
   }

   public boolean putGvalg(double glad, double glod, double val) {

      if(glad < glamn) return false;
      if(glad > glamx) return false;
      if(glod < glomn) return false;
      if(glod > glomx) return false;

      int ila=(int)Math.round((glad-glamn)/dgla)+1;     //*** nearest int
      int ilo=(int)Math.round((glod-glomn)/dglo)+1;

      g[ila][ilo] = val;
      return true;
   }

   public void readGrid(BufferedReader myBR) throws IOException, 
                                                    NumberFormatException {

//*** outer loop over latitude (south to north)
//*** inner loop over longitude (west to east)
//*** fortran-style index

      for(int ila=1; ila<=nla; ila++){
         for(int jlo=1; jlo<=nlo; jlo++){
            String myS = myBR.readLine();           //*** one value per line
            g[ila][jlo] = Double.parseDouble(myS.trim());
         }
      }//*** end outer loop
   }

//*** view functions

   public double getGglamn()  {return glamn;}
   public double getGglomn()  {return glomn;}
   public double getGglamx()  {return glamx;}
   public double getGglomx()  {return glomx;}
   public double getGdgla()   {return dgla;}
   public double getGdglo()   {return dglo;}
   public int    getGnla()    {return nla;}
   public int    getGnlo()    {return nlo;}

   public boolean inGrid(double glad, double glod) {
      if(glad < glamn) return false;
      if(glad > glamx) return false;
      if(glod < glomn) return false;
      if(glod > glomx) return false;
      return true;
   }

   public double getGval (int ila, int ilo) {return g[ila][ilo];}

   public double getGvalg(double glad, double glod) {

//*** look up -- nearest node

      if(glad < glamn) return GNULL;
      if(glad > glamx) return GNULL;
      if(glod < glomn) return GNULL;
      if(glod > glomx) return GNULL;

      int ila=(int)Math.round((glad-glamn)/dgla)+1;     //*** nearest int
      int ilo=(int)Math.round((glod-glomn)/dglo)+1;

      return g[ila][ilo];
   }

   public double blterpG(double glad, double glod) {

//*** compute bilinear interpolation
//*** logic not tested for '0/360 meridian wraparound'

      int    ix,jy,ix1,jy1;
      double x,y,a,b,c,d,w1,w2,w3,w4;

//*** grace region regarding boundary checks
//*** note: changes to lat/lon remain internal -- not passed back

      if(glad+BTOL < glamn) return GNULL;
      if(glad-BTOL > glamx) return GNULL;
      if(glod+BTOL < glomn) return GNULL;
      if(glod-BTOL > glomx) return GNULL;

      if(glad < glamn) glad=glamn;   //*** lines above guarantee grace region
      if(glad > glamx) glad=glamx;
      if(glod < glomn) glod=glomn;
      if(glod > glomx) glod=glomx;

//*** within grid boundaries, get indicies in the grid

      ix= (int) Math.floor( (glod-glomn)/dglo +1.0 );
      jy= (int) Math.floor( (glad-glamn)/dgla +1.0 );

//*** just being safe, should not trigger with boundary clipping

      if(ix < 1)   Dump.Prog("blterpG error1: ix= " +ix);
      if(jy < 1)   Dump.Prog("blterpG error2: jy= " +jy);
      if(ix > nlo) Dump.Prog("blterpG error3: ix= " +ix);
      if(jy > nla) Dump.Prog("blterpG error4: jy= " +jy);

//*** this logic pads the grid boundary on north and east edges
//*** useful when re-interpolating the grid

      ix1=ix+1;
      jy1=jy+1;
      if(ix1 > nlo) ix1=nlo;
      if(jy1 > nla) jy1=nla;

//*** now extract corners in order sw,se,nw,ne
//***                               c d
//***                               a b

      a=g[jy ][ix ];
      b=g[jy ][ix1];
      c=g[jy1][ix ];
      d=g[jy1][ix1];

      x=(glod-dglo*(ix-1)-glomn)/dglo;
      y=(glad-dgla*(jy-1)-glamn)/dgla;

//*** logic for imbedded nulls

      if(a <= NULL2 || b <= NULL2 || c <= NULL2 || d <= NULL2) {
        if(a <= NULL2 && b <= NULL2 && c <= NULL2 && d <= NULL2) return GNULL;

         w1 = shepwt(    x,     y, a);
         w2 = shepwt(1.0-x,     y, b);
         w3 = shepwt(    x, 1.0-y, c);
         w4 = shepwt(1.0-x, 1.0-y, d);

         return (a*w1+b*w2+c*w3+d*w4)/(w1+w2+w3+w4);

//*** no imbedded nulls -- standard bilinear interpolation

      }else{
         return a*(1.0-x)*(1.0-y)+
                b*     x *(1.0-y)+
                c*(1.0-x)*     y +
                d*     x *     y ;
      };
   }

   public double shepwt(double dx, double dy, double val) {

//*** compute a shepard's method weight

      if(val <= NULL2){ 
         return 0.0;            //*** no weight for imbedded nulls
      }else{
         double w = dx*dx+dy*dy;
         if(w < 0.000001) return 1000000.0;   //*** prevent divide by zero
         return 1.0/w;
      }
   }

   public double bqterpG(double glad, double glod) {

//*** compute biquadratic interpolation
//*** logic not tested for '0/360 meridian wraparound'

      int    ix,ix1,ix2,jy,jy1,jy2;
      double x,y,fx0,fx1,fx2;

//*** grace region regarding boundary checks
//*** note: changes to lat/lon remain internal -- not passed back

      if(glad+BTOL < glamn) return GNULL;
      if(glad-BTOL > glamx) return GNULL;
      if(glod+BTOL < glomn) return GNULL;
      if(glod-BTOL > glomx) return GNULL;

      if(glad < glamn) glad=glamn;   //*** lines above guarantee grace region
      if(glad > glamx) glad=glamx;
      if(glod < glomn) glod=glomn;
      if(glod > glomx) glod=glomx;

//*** within grid boundaries, get indicies in the grid

      ix= (int) Math.floor( (glod-glomn)/dglo +1.0 );
      jy= (int) Math.floor( (glad-glamn)/dgla +1.0 );

//*** just being safe, should not trigger with boundary clipping

      if(ix < 1)   Dump.Prog("bqterpG error1: ix= " +ix);
      if(jy < 1)   Dump.Prog("bqterpG error2: jy= " +jy);
      if(ix > nlo) Dump.Prog("bqterpG error3: ix= " +ix);
      if(jy > nla) Dump.Prog("bqterpG error4: jy= " +jy);

      ix1=ix+1;
      ix2=ix+2;

//*** back off if edge collision (do it twice)

      if(ix2 > nlo){
         ix =ix -1;
         ix1=ix1-1;
         ix2=ix2-1;
      }

      if(ix2 > nlo){
         ix =ix -1;
         ix1=ix1-1;
         ix2=ix2-1;
      }

      x=(glod-dglo*(ix-1)-glomn)/dglo;

//*** move grid to get nearer center of 3 x 3

      if( (x < 0.5) && (ix > 1) ){
         ix =ix -1;
         ix1=ix1-1;
         ix2=ix2-1;
         x  =x+1.0;
      }

      if( (x < 0.0) || (x > 2.0) ) Dump.Prog("bqterpG error5: x= " +x);

//*** now do it for y *************************************************

      jy1=jy+1;
      jy2=jy+2;
      if(jy2 > nla){
         jy =jy -1;
         jy1=jy1-1;
         jy2=jy2-1;
      }
      if(jy2 > nla){
         jy =jy -1;
         jy1=jy1-1;
         jy2=jy2-1;
      }
      y=(glad-dgla*(jy-1)-glamn)/dgla;
      if( (y < 0.5) && (jy > 1) ){
         jy =jy -1;
         jy1=jy1-1;
         jy2=jy2-1;
         y  =y+1.0;
      }
      if( (y < 0.0) || (y > 2.0+10*GTOL) ) Dump.Prog("bqterpG error6: y= " +y);

      fx0 =  qterp(x, g[jy ][ix ], g[jy ][ix1], g[jy ][ix2]);
      fx1 =  qterp(x, g[jy1][ix ], g[jy1][ix1], g[jy1][ix2]);
      fx2 =  qterp(x, g[jy2][ix ], g[jy2][ix1], g[jy2][ix2]);
      return qterp(y, fx0        , fx1        , fx2        );
   }

   private double qterp(double x, double f0, double f1, double f2) {

//*** linear quadratic interpolation from equally spaced values
//*** uses newton-gregory forward polynomial
//*** x ranges from 0 thru 2.  (thus s = x)

//*** forward differences

      double df0  = f1 -f0;
      double df1  = f2 -f1;
      double d2f0 = df1-df0;

      return  f0 + x*df0 + 0.5*x*(x-1.0)*d2f0;
   }

}//***endclass GPos
//************************************************************************
class Dump {

//*** Fatal Error class -- Dump

   public static void Fatal(String message) {

//*** for end-user consumption

      JOptionPane.showMessageDialog(null, message, "Fatal Error", 
            JOptionPane.ERROR_MESSAGE); 
      System.exit(1);
   } 

   public static void Prog(String message) {

//*** these should never occur

      JOptionPane.showMessageDialog(null, message, 
           "Internal Programming Error -- Fatal", 
            JOptionPane.ERROR_MESSAGE); 
      System.exit(1);
   } 
}//***endclass Dump
//************************************************************************
